;;; ----------------------
;;; Normalized polynomials
;;; ----------------------

(in-package "NPOL")

(include-book "upol")
(include-book "ordinal-ordering")

;;; NOTE:
;;;
;;; From now on we won't assume that the ordinal embedding of a
;;; term/monomial is not zero. Instead, it will be shown that it is
;;; always possible to build a modified ordinal embedding holding this
;;; property without imposing any additional constraint. This approach
;;; has the advantage of separating two concerns: the development of
;;; monomial orderings and the development of induced polynomial
;;; orderings.

(in-theory (disable TER::|~(term->e0-ordinal(a) = 0)|
		    MON::|~(monomial->e0-ordinal(a) = 0)|))

;;; ---------
;;; Functions
;;; ---------

;;; Recognizer.

(defun polynomialp (p)
  (declare (xargs :guard t))
  (and (UPOL::polynomialp p) (UPOL::nfp p)))

;;; ...

;;; Polynomial ordering.

(defun < (p q)
  (declare (xargs :guard (and (polynomialp p) (polynomialp q))))
  (cond ((or (endp p) (endp q))
         (not (endp q)))
	((MON::=T (first p) (first q))
         (< (rest p) (rest q)))
        (t
         (MON::< (first p) (first q)))))

;;; Ordinal increment.

(defun ordinal-increment (a)
  (declare (xargs :guard t))
  (if (consp a)
      (if (and (atom (rest a)) (integerp (rest a)))
	  (cons (first a) (+ (rest a) 1))
	(cons (first a) (ordinal-increment (rest a))))
    (if (and (integerp a) (<= 0 a))
	(+ a 1)
      a)))

;;; Modified monomial ordinal embedding.

(defun modified-monomial->e0-ordinal (a)
  (declare (xargs :guard (consp a)))
  (ordinal-increment (monomial->e0-ordinal a)))

;;; Polynomial ordinal embedding.

(defun polynomial->e0-ordinal (p)
  (declare (xargs :guard (polynomialp p)))
  (if (endp p)
      0
    (cons (modified-monomial->e0-ordinal (first p))
          (polynomial->e0-ordinal (rest p)))))

;;; ----------
;;; Properties
;;; ----------

;;; ...

;;; The polynomial ordering is a partial strict order.

(defthm |~(p < p)|
  (not (< p p)))

(defthm |p < q & q < r => p < r|
  (implies (and (< p q) (< q r))
	   (< p r)))

;;; Antisymmetry of the polynomial ordering.

(defthm |p < q => ~(q < p)|
  (implies (< p q)
	   (not (< q p)))
  :hints (("Goal"
	   :in-theory (disable |p < q & q < r => p < r|)
	   :use (:instance |p < q & q < r => p < r| (r p)))))

;;; The ordinal increment of an ordinal is an ordinal.

(defthm |e0-ordinalp(a) => e0-ordinalp(ordinal-increment(a)|
  (implies (e0-ordinalp a)
	   (e0-ordinalp (ordinal-increment a))))

;;; The ordinal increment does not produce 0.

(defthm |~(ordinal-increment(a) = 0)|
  (not (equal (ordinal-increment a) 0)))

;;; If an ordinal is less than other, so their ordinal increments are.

;;; NOTE:
;;;
;;; The proof of this theorem is extremely complex. We are amazed at
;;; how ACL2 can prove this on its own.

(defthm |a <e0 b => a + 1 <e0 b + 1|
  (implies (and (e0-ordinalp a) (e0-ordinalp b)
		(e0-ord-< a b))
	   (e0-ord-< (ordinal-increment a) (ordinal-increment b))))

;;; Congruence. 

(defcong MON::=T equal (modified-monomial->e0-ordinal a) 1)

;;; The monomial ordinal embedding does not produce 0.

(defthm |~(modified-monomial->e0-ordinal(a) = 0)|
  (not (equal (modified-monomial->e0-ordinal a) 0)))

;;; Well-foundedness of the monomial ordering w.r.t. the modified
;;; monomial ordinal embedding.

(defthm modified-well-foundedness-of-<
  (and (implies (monomialp a)
		(e0-ordinalp (modified-monomial->e0-ordinal a)))
       (implies (and (monomialp a) (monomialp b)
		     (MON::< a b))
		(e0-ord-< (modified-monomial->e0-ordinal a)
			  (modified-monomial->e0-ordinal b))))
  :rule-classes (:rewrite :well-founded-relation))

;;; Correctness of the polynomial ordinal embedding.

(defthm correctness-of-polynomial->e0-ordinal
  (implies (polynomialp p)
           (e0-ordinalp (polynomial->e0-ordinal p))))

;;; Well-foundedness of the polynomial ordering.

(defthm well-foundedness-of-<
  (and (implies (polynomialp p)
                (e0-ordinalp (polynomial->e0-ordinal p)))
       (implies (and (polynomialp p) (polynomialp q)
                     (< p q))
                (e0-ord-< (polynomial->e0-ordinal p)
                          (polynomial->e0-ordinal q))))
  :rule-classes :well-founded-relation)
